@testset "Transforms" begin
  @testset "Builtin" begin
    # transforms that are revertible
    d = georef((z=rand(100), w=rand(100)))
    for p in [
      Select(:z),
      Reject(:z),
      Rename(:z => :a),
      StdNames(),
      Sort(:z),
      Sample(10),
      Filter(x -> true),
      DropMissing(),
      Replace(1.0 => 2.0),
      Coalesce(value=0.0),
      Coerce(:z => ST.Continuous),
      Identity(),
      Center(),
      Scale(),
      MinMax(),
      Interquartile(),
      ZScore(),
      Quantile(),
      Functional(sin),
      EigenAnalysis(:V),
      PCA(),
      DRS(),
      SDS(),
      RowTable(),
      ColTable()
    ]
      n, c = apply(p, d)
      t = Tables.columns(n)
      r = revert(p, n, c)
      @test n isa AbstractGeoTable
      @test r isa AbstractGeoTable
    end

    # transforms with categorical variables
    d = georef((c=categorical([1, 2, 3]),))
    for p in [Levels(:c => [1, 2, 3]), OneHot(:c)]
      n, c = apply(p, d)
      t = Tables.columns(n)
      r = revert(p, n, c)
      @test n isa AbstractGeoTable
      @test r isa AbstractGeoTable
    end

    d = georef((z=rand(100), w=rand(100)))
    p = Select(:w)
    n, c = apply(p, d)
    t = Tables.columns(n)
    @test Tables.columnnames(t) == (:w, :geometry)

    d = georef((z=rand(100), w=rand(100)))
    p = Sample(100)
    n, c = apply(p, d)
    r = revert(p, n, c)
    @test r == d
    t = Tables.columns(n)
    @test Tables.columnnames(t) == (:z, :w, :geometry)

    d = georef((a=[1, missing, 3], b=[3, 2, 1]))
    p = DropMissing()
    n, c = apply(p, d)
    @test Tables.columns(values(n)) == (a=[1, 3], b=[3, 1])
    @test nelements(domain(n)) == 2
    r = revert(p, n, c)
    @test r == d
  end

  @testset "Geometric" begin
    d = georef((z=rand(100), w=rand(100)))
    p = StdCoords()
    n, c = apply(p, d)
    dom = domain(n)
    cen = centroid.(dom)
    xs = first.(coordinates.(cen))
    @test dom isa CartesianGrid
    @test all(x -> -0.5 ≤ x ≤ 0.5, xs)
  end

  @testset "Detrend" begin
    rng = MersenneTwister(42)
    l = range(-1, stop=1, length=100)
    μ = [x^2 + y^2 for x in l, y in l]
    ϵ = 0.1rand(rng, 100, 100)
    d = georef((z=μ + ϵ, w=rand(100, 100)))
    p = Detrend(:z, degree=2)
    n, c = apply(p, d)
    r = revert(p, n, c)
    D = Tables.matrix(values(d))
    R = Tables.matrix(values(r))
    @test isapprox(D, R, atol=1e-6)
  end

  @testset "Potrace" begin
    # challenging case with letters
    img = load(joinpath(datadir, "letters.png"))
    gtb = georef((color=img,))
    trans = Potrace(1)
    ngtb, cache = apply(trans, gtb)
    ndom = domain(ngtb)
    @test nelements(ndom) == 2
    @test eltype(ndom) <: Multi
    polys1 = parent(ndom[1])
    polys2 = parent(ndom[2])
    @test length(polys1) == 4
    @test length(polys2) == 2
    rgtb = revert(trans, ngtb, cache)
    dom = domain(gtb)
    rdom = domain(rgtb)
    @test rdom isa Grid
    @test size(rdom) == size(dom)
    @test minimum(rdom) == minimum(dom)
    @test maximum(rdom) == maximum(dom)
    @test spacing(rdom) == spacing(dom)

    # concentric circles
    ball1 = Ball((0, 0), 1)
    ball2 = Ball((0, 0), 2)
    ball3 = Ball((0, 0), 3)
    grid = CartesianGrid((-5, -5), (5, 5), dims=(100, 100))
    inds1 = centroid.(grid) .∈ Ref(ball1)
    inds2 = centroid.(grid) .∈ Ref(ball2)
    inds3 = centroid.(grid) .∈ Ref(ball3)
    mask = zeros(100, 100)
    mask[inds3] .= 1
    mask[inds2] .= 0
    mask[inds1] .= 1
    dat = georef((mask=mask,))
    new = dat |> Potrace(1)
    dom = domain(new)
    @test nelements(dom) == 2
    @test eltype(dom) <: Multi
    polys1 = parent(dom[1])
    polys2 = parent(dom[2])
    @test length(polys1) == 2
    @test length(polys2) == 2
    new1 = dat |> Potrace(1, ϵ=0.1)
    new2 = dat |> Potrace(1, ϵ=0.5)
    dom1 = domain(new1)
    dom2 = domain(new2)
    for (g1, g2) in zip(dom1, dom2)
      @test nvertices(g1) > nvertices(g2)
    end

    # make sure that aggregation works
    Z = [sin(i / 10) + sin(j / 10) for i in 1:100, j in 1:100]
    M = Z .> 0
    Ω = georef((Z=Z, M=M))
    𝒯 = Ω |> Potrace(:M, :Z => mean)
    masks = unique(Ω.M)
    @test nelements(domain(𝒯)) == 2
    @test Set(𝒯.M) == Set([true, false])
    @test all(z -> -1 ≤ z ≤ 1, 𝒯.Z)
    @test 𝒯.Z[1] == mean(Ω.Z[masks[1] .== Ω.M])
    @test 𝒯.Z[2] == mean(Ω.Z[masks[2] .== Ω.M])
  end

  @testset "Rasterize" begin
    @test isrevertible(Rasterize(10, 10)) == true

    x = [1, 2, 3, 4, 5]
    y = [1.1, 2.2, 3.3, 4.4, 5.5]
    pts = Point2[(3, 9), (7, 8), (8, 5), (5, 4), (1, 5)]
    seg1 = Segment(pts[1], pts[2])
    seg2 = Segment(pts[2], pts[3])
    seg3 = Segment(pts[3], pts[4])
    seg4 = Segment(pts[4], pts[5])
    seg5 = Segment(pts[5], pts[1])
    poly1 = PolyArea((2, 0), (6, 2), (2, 2))
    poly2 = PolyArea((0, 6), (3, 8), (0, 10))
    poly3 = PolyArea((3, 6), (9, 6), (9, 9), (6, 9))
    poly4 = PolyArea((7, 0), (10, 0), (10, 4), (7, 4))
    poly5 = PolyArea((1, 3), (5, 3), (6, 6), (3, 8), (0, 6))

    gtb = georef((; x, y), pts)
    grid = CartesianGrid(10, 10)
    trans = Rasterize(grid)
    ngtb, cache = apply(trans, gtb)
    linds = LinearIndices((10, 10))
    @test ngtb.x[linds[3, 9]] == 1
    @test ngtb.x[linds[7, 8]] == 2
    @test ngtb.x[linds[8, 5]] == 3
    @test ngtb.x[linds[5, 4]] == 4
    @test ngtb.x[linds[1, 5]] == 5
    @test ngtb.y[linds[3, 9]] == 1.1
    @test ngtb.y[linds[7, 8]] == 2.2
    @test ngtb.y[linds[8, 5]] == 3.3
    @test ngtb.y[linds[5, 4]] == 4.4
    @test ngtb.y[linds[1, 5]] == 5.5

    gtb = georef((; x, y), [seg1, seg2, seg3, seg4, seg5])
    grid = CartesianGrid((0, 0), (10, 10), dims=(20, 20))
    trans = Rasterize(grid)
    ngtb, cache = apply(trans, gtb)
    linds = LinearIndices((20, 20))
    @test ngtb.x[linds[10, 17]] == 1
    @test ngtb.x[linds[15, 13]] == 2
    @test ngtb.x[linds[13, 9]] == 3
    @test ngtb.x[linds[6, 9]] == 4
    @test ngtb.x[linds[4, 14]] == 5
    @test ngtb.y[linds[10, 17]] == 1.1
    @test ngtb.y[linds[15, 13]] == 2.2
    @test ngtb.y[linds[13, 9]] == 3.3
    @test ngtb.y[linds[6, 9]] == 4.4
    @test ngtb.y[linds[4, 14]] == 5.5

    gtb = georef((; x, y), [poly1, poly2, poly3, poly4, poly5])
    trans = Rasterize(20, 20, :x => last, :y => std)
    ngtb, cache = apply(trans, gtb)
    linds = LinearIndices((20, 20))
    @test ngtb.x[linds[7, 3]] == 1
    @test ngtb.x[linds[3, 16]] == 2
    @test ngtb.x[linds[15, 15]] == 3
    @test ngtb.x[linds[17, 5]] == 4
    @test ngtb.x[linds[6, 11]] == 5
    @test ngtb.y[linds[7, 3]] == 1.1
    @test ngtb.y[linds[3, 16]] == 2.2
    @test ngtb.y[linds[15, 15]] == 3.3
    @test ngtb.y[linds[17, 5]] == 4.4
    @test ngtb.y[linds[6, 11]] == 5.5
    # intersection: poly3 with poly5
    @test ngtb.x[linds[9, 13]] == last(gtb.x[[3, 5]])
    @test ngtb.y[linds[9, 13]] == std(gtb.y[[3, 5]])

    # revert
    gtb = georef((; z=1:4), [poly1, poly2, poly3, poly4])
    trans = Rasterize(200, 200)
    ngtb, cache = apply(trans, gtb)
    rgtb = revert(trans, ngtb, cache)
    inds = filter(!iszero, unique(cache))
    @test isapprox(area(gtb.geometry[inds[1]]), area(rgtb.geometry[1]), atol=0.5)
    @test isapprox(area(gtb.geometry[inds[2]]), area(rgtb.geometry[2]), atol=0.5)
    @test isapprox(area(gtb.geometry[inds[3]]), area(rgtb.geometry[3]), atol=0.5)
    @test isapprox(area(gtb.geometry[inds[4]]), area(rgtb.geometry[4]), atol=0.5)
    # geotable with "mask" column
    gtb = georef((; z=1:4, mask=4:-1:1), [poly1, poly2, poly3, poly4])
    trans = Rasterize(10, 10)
    ngtb, cache = apply(trans, gtb)
    rgtb = revert(trans, ngtb, cache)
    @test nrow(rgtb) == nrow(gtb)
    @test ncol(rgtb) == ncol(gtb)
    @test propertynames(rgtb) == propertynames(gtb)
  end

  @testset "UniqueCoords" begin
    @test isrevertible(UniqueCoords()) == false

    X = [i * j for i in 1:2, j in 1:1_000_000]
    z = rand(1_000_000)
    d = georef((z=[z; z],), [X X])
    u = d |> UniqueCoords()
    du = domain(u)
    p = [centroid(du, i) for i in 1:nelements(du)]
    U = reduce(hcat, coordinates.(p))
    @test nelements(du) == 1_000_000
    @test Set(eachcol(U)) == Set(eachcol(X))

    X = rand(3, 100)
    z = rand(100)
    n = [string(i) for i in 1:100]
    Xd = hcat(X, X[:, 1:10])
    zd = vcat(z, z[1:10])
    nd = vcat(n, n[1:10])
    sdata = georef((z=zd, n=nd), PointSet(Xd))
    ndata = sdata |> UniqueCoords()
    @test nrow(ndata) == 100

    # domain with repeated points
    x = rand(100)
    y = rand(1:10, 100)
    table = (; x, y)
    points = Point.(1:10, 10:-1:1)

    pset = PointSet(rand(points, 100))
    sdata = georef(table, pset)
    ndata = sdata |> UniqueCoords()
    @test nrow(ndata) == 10

    # aggregators
    pset = PointSet(repeat(points, inner=10))
    sdata = georef(table, pset)

    # default aggregators
    ndata = sdata |> UniqueCoords()
    @test nrow(ndata) == 10

    for i in 1:10
      j = i * 10
      @test ndata.x[i] == mean(sdata.x[(j - 9):j])
    end

    for i in 1:10
      j = i * 10
      @test ndata.y[i] == first(sdata.y[(j - 9):j])
    end

    # custom aggregators
    # colspec: index
    ndata = sdata |> UniqueCoords(1 => std, 2 => median)
    @test nrow(ndata) == 10

    for i in 1:10
      j = i * 10
      @test ndata.x[i] == std(sdata.x[(j - 9):j])
    end

    for i in 1:10
      j = i * 10
      @test ndata.y[i] == median(sdata.y[(j - 9):j])
    end

    # colspec: symbols
    ndata = sdata |> UniqueCoords(:x => last, :y => first)
    @test nrow(ndata) == 10

    for i in 1:10
      j = i * 10
      @test ndata.x[i] == last(sdata.x[(j - 9):j])
    end

    for i in 1:10
      j = i * 10
      @test ndata.y[i] == first(sdata.y[(j - 9):j])
    end

    # colspec: strings
    ndata = sdata |> UniqueCoords("x" => maximum, "y" => minimum)
    @test nrow(ndata) == 10

    for i in 1:10
      j = i * 10
      @test ndata.x[i] == maximum(sdata.x[(j - 9):j])
    end

    for i in 1:10
      j = i * 10
      @test ndata.y[i] == minimum(sdata.y[(j - 9):j])
    end
  end

  @testset "CoDa" begin
    d = georef((z=rand(1000), w=rand(1000)))

    n = d |> Closure()
    t = Tables.columns(n)
    @test n isa AbstractGeoTable
    @test Tables.columnnames(t) == (:z, :w, :geometry)

    n = d |> Remainder()
    t = Tables.columns(n)
    @test n isa AbstractGeoTable
    @test Tables.columnnames(t) == (:z, :w, :remainder, :geometry)

    n = d |> ALR()
    t = Tables.columns(n)
    @test n isa AbstractGeoTable
    @test Tables.columnnames(t) == (:z, :geometry)

    n = d |> CLR()
    t = Tables.columns(n)
    @test n isa AbstractGeoTable
    @test Tables.columnnames(t) == (:z, :w, :geometry)

    n = d |> ILR()
    t = Tables.columns(n)
    @test n isa AbstractGeoTable
    @test Tables.columnnames(t) == (:z, :geometry)
  end

  @testset "Mixed" begin
    d = georef((z=rand(1000), w=rand(1000)))
    p = Quantile() → StdCoords()
    n, c = apply(p, d)
    r = revert(p, n, c)
    Xr = Tables.matrix(values(r))
    Xd = Tables.matrix(values(d))
    @test isapprox(Xr, Xd, atol=0.1)
  end
end
